#include "RandomSaver.h"
#include <interface/StringView.h>
#include <kernel/image.h>
#include <storage/Directory.h>
#include <storage/Entry.h>
#include <storage/File.h>
#include <storage/Path.h>
#include <stdlib.h>
#include <stdio.h>

#define ArpDEBUG 0

#if ArpDEBUG
#define D(x) x
#else
#define D(x)
#endif

// --------------------- MAIN INSTANTIATION FUNCTION ---------------------

extern "C" _EXPORT BScreenSaver *instantiate_screen_saver(BMessage *message, image_id image)
{
	return new RandomSaver(message, image);
}

// --------------------- MISC SUPPORT FUNCTIONS ---------------------

static status_t make_relative_path(directory_which relFrom, const char* relTo,
								   BPath* path)
{
	if( !path ) return B_OK;
	
	status_t err = find_directory(relFrom,path);
	if( err == B_OK && relTo != 0 ) {
		err = path->Append(relTo);
	}
	return err;
}

static status_t open_relative_file(directory_which relFrom, const char* relTo,
								   BFile* file)
{
	BPath filePath;
	status_t err = make_relative_path(relFrom, relTo, &filePath);
	if( err == B_OK && file ) {
		file->SetTo(filePath.Path(), B_READ_ONLY);
	}
	return err;
}

static status_t open_relative_dir(directory_which relFrom, const char* relTo,
								  BDirectory* dir, BPath* dirPath=0)
{
	BPath filePath;
	if( !dirPath ) dirPath = &filePath;
	status_t err = make_relative_path(relFrom, relTo, dirPath);
	if( err == B_OK && dir ) {
		dir->SetTo(dirPath->Path());
	}
	return err;
}

typedef BScreenSaver* (ScreenSaverFunc)(BMessage* msg, image_id id);

static image_id load_relative_add_on(directory_which base,
									 const char* path, const char* name)
{
	BPath filePath;
	status_t err = make_relative_path(base, path, &filePath);
	if( err == B_OK && name ) {
		err = filePath.Append(name, true);
	}
	if( err == B_OK ) {
		return load_add_on(filePath.Path());
	}
	return (image_id)err;
}

// --------------------- RANDOM SAVER IMPLEMENTATION ---------------------

RandomSaver::RandomSaver(BMessage *message, image_id image)
		: BScreenSaver(message, image),
		  mMyImage(image), mMyName(""),
		  mMySettings(*message), mRealSaver(0),
		  mRealImage(-1), mIsPreview(true)
{
	D(printf("Creating a new saver: %p\n", this));
	srand((unsigned int)system_time());
	
	image_info info;
	if( get_image_info(mMyImage, &info) == B_OK ) {
		BPath path(info.name);
		mMyName = path.Leaf();
		D(printf("Name of my saver: %s\n", mMyName.String()));
	}
}

RandomSaver::~RandomSaver()
{
	D(printf("Deleting my saver: %p\n", this));
	UnloadSaver();
}

status_t RandomSaver::InitCheck()
{
	// Do something evil: poke into ScreenSaver's settings file.
	BFile settingsFile;
	status_t err = open_relative_file(B_USER_SETTINGS_DIRECTORY,
									  "ScreenSaver_settings",
									  &settingsFile);
	if( err != B_OK ) {
		err = open_relative_file(B_COMMON_SETTINGS_DIRECTORY,
								 "ScreenSaver_settings",
								 &settingsFile);
	}
	if( err == B_OK ) {
		err = mSaverSettings.Unflatten(&settingsFile);
	}
	
	BMessage availMods;
	AddModules(B_USER_ADDONS_DIRECTORY, "Screen Savers", &availMods);
	AddModules(B_COMMON_ADDONS_DIRECTORY, "Screen Savers", &availMods);
	AddModules(B_BEOS_ADDONS_DIRECTORY, "Screen Savers", &availMods);
	
	bool initialized=false;
	mMySettings.FindBool("initialized", &initialized);
	if( !initialized ) {
		
		mMySettings = availMods;
		mMySettings.AddBool("initialized", true);
		int32 i=0;
		const char* name=0;
		while( mMySettings.FindString("name", i, &name) == B_OK ) {
			mMySettings.AddString("sel", name);
			i++;
		}
		
	} else {
		// Synchronize found available modules with list of
		// selected modules.
		int32 i=0;
		const char* name=0;
		
		D(printf("Looking for new modules...\n"));
		while( availMods.FindString("name", i, &name) == B_OK ) {
			int32 j=0;
			const char* has=0;
			while( mMySettings.FindString("name", j, &has) == B_OK ) {
				if( strcmp(name, has) == 0 ) break;
				has = 0;
				j++;
			}
			if( !has ) {
				D(printf("Adding new module: %s\n", name));
				mMySettings.AddString("name", name);
				mMySettings.AddString("sel", name);
			}
			i++;
		}
		
		i=0;
		
		D(printf("Looking for removed modules...\n"));
		while( mMySettings.FindString("name", i, &name) == B_OK ) {
			int32 j=0;
			const char* has=0;
			while( availMods.FindString("name", j, &has) == B_OK ) {
				if( strcmp(name, has) == 0 ) break;
				has = 0;
				j++;
			}
			if( !has ) {
				D(printf("Removing lost module: %s\n", name));
				mMySettings.RemoveData("name", i);
			} else {
				i++;
			}
		}
		
		i=0;
		
		D(printf("Looking for removed selected modules...\n"));
		while( mMySettings.FindString("sel", i, &name) == B_OK ) {
			int32 j=0;
			const char* has=0;
			while( availMods.FindString("name", j, &has) == B_OK ) {
				if( strcmp(name, has) == 0 ) break;
				has = 0;
				j++;
			}
			if( !has ) {
				D(printf("Removing lost selected module: %s\n", name));
				mMySettings.RemoveData("sel", i);
			} else {
				i++;
			}
		}
		
	}
	
	return err;
}

status_t RandomSaver::StartSaver(BView *view, bool preview)
{
	D(printf("Starting saver, preview=%d\n", preview));
	
	mIsPreview = preview;
	status_t err = B_OK;
	
	if( !mIsPreview ) {
		err = LoadRandomSaver();
		if( err != B_OK ) return err;
		if( mRealSaver ) {
			D(printf("Telling child saver to start.\n"));
			mIsPreview = preview;
			err = mRealSaver->StartSaver(view, preview);
			CopyChildState();
			D(printf("Result of start is %ld\n", err));
		}
	}
	
	return err;
}

void RandomSaver::StopSaver()
{
	if( !mIsPreview && mRealSaver ) {
		D(printf("Stopping child saver.\n"));
		mRealSaver->StopSaver();
		CopyChildState();
		UnloadSaver();
	}
}

void RandomSaver::Draw(BView *view, int32 frame)
{
	if( !mIsPreview && mRealSaver ) {
		//D(printf("Drawing child saver.\n"));
		mRealSaver->Draw(view, frame);
		CopyChildState();
	}
}

void RandomSaver::DirectConnected(direct_buffer_info* info)
{
	if( !mIsPreview && mRealSaver ) {
		D(printf("Direct connecting child saver.\n"));
		mRealSaver->DirectConnected(info);
		CopyChildState();
	}
}

void RandomSaver::DirectDraw(int32 frame)
{
	if( !mIsPreview && mRealSaver ) {
		//D(printf("Direct drawing child saver.\n"));
		mRealSaver->DirectDraw(frame);
		CopyChildState();
	}
}

void RandomSaver::StartConfig(BView *view)
{
	view->AddChild(new BStringView(BRect(10, 10, 200, 35), B_EMPTY_STRING, "Random Saver, playing with fate."));
}

void RandomSaver::StopConfig()
{
}

void RandomSaver::SupplyInfo(BMessage* info) const
{
	if( !mIsPreview && mRealSaver ) {
		D(printf("Supplying info from child saver.\n"));
		mRealSaver->SupplyInfo(info);
	}
}

void RandomSaver::ModulesChanged(const BMessage* info)
{
	if( !mIsPreview && mRealSaver ) {
		D(printf("Modules changed for child saver.\n"));
		mRealSaver->ModulesChanged(info);
		CopyChildState();
	}
}

status_t RandomSaver::SaveState(BMessage* info) const
{
	*info = mMySettings;
	return B_OK;
}

status_t RandomSaver::AddModules(directory_which path, const char* file,
								 BMessage* to) const
{
	if( !to ) return B_OK;
	
	BDirectory moduleDir;
	BPath dirPath;
	status_t err = open_relative_dir(path, file, &moduleDir, &dirPath);
	if( err == B_OK ) {
		BEntry entry;
		moduleDir.Rewind();
		while( moduleDir.GetNextEntry(&entry) == B_OK ) {
			if( !entry.IsDirectory() ) {
				char name[B_FILE_NAME_LENGTH];
				if( entry.GetName(name) == B_OK ) {
					BPath path(dirPath.Path(), name);
					image_id image = load_add_on(path.Path());
					if( image >= 0 ) {
						if( mMyName != path.Leaf() ) {
							to->AddString("name", path.Leaf());
						}
						unload_add_on(image);
					}
				}
			}
		}
	}
	
	return err;
}

status_t RandomSaver::LoadRandomSaver()
{
	UnloadSaver();
	
	int32 which=0;
	type_code type=0;
	mMySettings.GetInfo("sel", &type, &which);
	if( which <= 0 ) return B_ERROR;
	
	D(printf("Found %ld modules.  Random=%d\n", which, rand()));
	which = rand()%which;
	const char* modName = 0;
	mMySettings.FindString("sel", which, &modName);
	D(printf("Selected module #%ld: %s.\n", which, modName));
	if( !modName ) return B_ERROR;
	
	D(printf("Loading module %s...\n", modName));
	mRealImage = load_relative_add_on(B_USER_ADDONS_DIRECTORY,
									  "Screen Savers", modName);
	if( mRealImage < 0 ) {
		mRealImage = load_relative_add_on(B_COMMON_ADDONS_DIRECTORY,
										  "Screen Savers", modName);
	}
	if( mRealImage < 0 ) {
		mRealImage = load_relative_add_on(B_BEOS_ADDONS_DIRECTORY,
										  "Screen Savers", modName);
	}
	
	D(printf("Image = %ld\n", mRealImage));
	if( mRealImage < 0 ) return (status_t)mRealImage;
	
	ScreenSaverFunc* func;
	status_t res;
	D(printf("Finding symbol...\n"));
	if( (res=get_image_symbol(mRealImage,"instantiate_screen_saver",
							  B_SYMBOL_TYPE_TEXT,(void**)&func)) == B_OK ) {
		
		D(printf("Symbol address is %p\n", func));
		BMessage settings;
		BString settingsName("modulesettings_");
		settingsName << modName;
		D(printf("Trying to find settings in %s\n", settingsName.String()));
		mSaverSettings.FindMessage(settingsName.String(), &settings);
		mRealSaver = (*func)(&settings, mRealImage);
		D(printf("Allocated screen saver is %p\n", mRealSaver));
		
		if( mRealSaver ) res = mRealSaver->InitCheck();
	}
	
	D(printf("Result of load=%ld, saver=%p\n", res, mRealSaver));
	
	if( res != B_OK || !mRealSaver ) {
		UnloadSaver();
		if( res == B_OK ) res = B_ERROR;
	}
	
	return res;
}

void RandomSaver::UnloadSaver()
{
	D(printf("Unloading saver.\n"));
	if( mRealSaver ) {
		mRealSaver->StopSaver();
		delete mRealSaver;
		mRealSaver = 0;
	}
	if( mRealImage >= 0 ) {
		unload_add_on(mRealImage);
		mRealImage = -1;
	}
	
	mIsPreview = true;
}

void RandomSaver::CopyChildState()
{
	if( !mIsPreview && mRealSaver ) {
		SetTickSize(mRealSaver->TickSize());
		SetLoop(mRealSaver->LoopOnCount(), mRealSaver->LoopOffCount());
	}
}
